//----------------------------------------------------------------------------------------
// This is an all-in-one sketch for a 32x32 matrix display (similar to Adafruit version).
//
// Selectable options are:
//    Game of Life display
//    Snowflake display
//    Snake game
//    Pong game
//    Tetris game
//
// Written by Boomer48: January 2021
//
//

#define R0 2
#define G0 3
#define B0 4
#define R1 5
#define G1 6
#define B1 7
#define CLK  8
#define OE   9
#define LAT 10
#define A   A0
#define B   A1
#define C   A2
#define D   A3

// Port definitions for switches and potentiometers
#define Left_Sw 11 // Snake and Tetris
#define Right_Sw 12 // Snake and Tetris
#define Up_Sw A4 // Snake
#define Down_Sw A5 // Snake
#define Rotate_Sw1 A4 // Tetris (same as Up_Sw)
#define Rotate_Sw2 A5 // Tetris (same as Down_Sw)
#define Paddle1_Port A6 // Pong
#define Paddle2_Port A7 // Pong

#define WIDTH 32
#define HEIGHT 32

#define RED 0
#define GRN 1
#define BLU 2
#define YEL 3
#define MAG 4
#define CYN 5
byte color;

unsigned long int cells[HEIGHT]; // 32 bits (columns) by HEIGHT bytes (rows)
byte column;
byte row;

byte Cycle_Timer = 0;
byte cycle_time; // Number of matrix update cycles between new generations

#define Life_Display 0
#define Snowflake_Display 1
#define Snake_Game 2
#define Pong_Game 3
#define Tetris_Game 4
byte Game_Num = Life_Display;
char ASCII[] = "0123456789ABCDEF";

// Make visible to include files
void Update_Matrix();

#include "Display_Characters.h"
#include "Snowflake_Display.h"
#include "Snake_Game.h"
#include "Pong_Game.h"
#include "Tetris_Game.h"

//-----------------------------------------------------------------------------
void setup() {
  byte temp = 0;
  unsigned long Old_Time;

  pinMode(R0, OUTPUT);
  pinMode(G0, OUTPUT);
  pinMode(B0, OUTPUT);
  pinMode(R1, OUTPUT);
  pinMode(G1, OUTPUT);
  pinMode(B1, OUTPUT);
  pinMode(CLK, OUTPUT);
  pinMode(OE, OUTPUT);
  pinMode(LAT, OUTPUT);
  pinMode(A, OUTPUT);
  pinMode(B, OUTPUT);
  pinMode(C, OUTPUT);
  pinMode(D, OUTPUT);

  digitalWrite(R0, LOW);
  digitalWrite(G0, LOW);
  digitalWrite(B0, LOW);
  digitalWrite(R1, LOW);
  digitalWrite(G1, LOW);
  digitalWrite(B1, LOW);
  digitalWrite(CLK, LOW);
  digitalWrite(OE, HIGH);
  digitalWrite(LAT, LOW);
  digitalWrite(A, LOW);
  digitalWrite(B, LOW);
  digitalWrite(C, LOW);
  digitalWrite(D, LOW);

  digitalWrite(Left_Sw, HIGH); // turn on internal pullup
  digitalWrite(Right_Sw, HIGH); // turn on internal pullup
  digitalWrite(Up_Sw, HIGH); // turn on internal pullup
  digitalWrite(Down_Sw, HIGH); // turn on internal pullup

  Old_Time = millis();

  while ((millis() - Old_Time) < 3000)
  {
    switch (Game_Num)
    {
      case Life_Display:
        Display_String("LIFE ", 15);
        break;
      case Snowflake_Display:
        Display_String("SNOW ", 15);
        break;
      case Snake_Game:
        Display_String("SNAKE", 15);
        break;
      case Pong_Game:
        Display_String("PONG ", 15);
        break;
      case Tetris_Game:
        Display_String("TETRS", 15);
        break;
    }

    Update_Matrix();

    if ((digitalRead(Left_Sw) == LOW) ||
        (digitalRead(Right_Sw) == LOW) ||
        (digitalRead(Up_Sw) == LOW) ||
        (digitalRead(Down_Sw) == LOW))
    {
      Game_Num = (Game_Num + 1) % 5;
      Old_Time = millis();

      // Clear matrix
      for (row = 0; row < HEIGHT; row++)
        cells[row] = 0;

      while ((digitalRead(Left_Sw) == LOW) ||
        (digitalRead(Right_Sw) == LOW) ||
        (digitalRead(Up_Sw) == LOW) ||
        (digitalRead(Down_Sw) == LOW))
        delay(50);
    }
  }

  // Clear matrix
  for (row = 0; row < HEIGHT; row++)
    cells[row] = 0;

  // Determine which display or game based on switch settings
  switch (Game_Num)
  {
    case Life_Display:
      Snowflake_setup();
      break;
    case Snowflake_Display:
      Snowflake_setup();
      break;
    case Snake_Game:
      Snake_setup();
      break;
    case Pong_Game:
      Pong_setup();
      break;
    case Tetris_Game:
      Tetris_setup();
      break;
  }
}

void loop() {
  // Run selected display or game
  switch (Game_Num)
  {
    case Life_Display:
      Snowflake_loop();
      break;
    case Snowflake_Display:
      Snowflake_loop();
      break;
    case Snake_Game:
      Snake_loop();
      break;
    case Pong_Game:
      Pong_loop();
      break;
    case Tetris_Game:
      Tetris_loop();
      break;
  }
}

void Update_Matrix() {
  unsigned long int temp;

  row = 0;

  while (row < HEIGHT / 2)
  {
    digitalWrite(A, bitRead(row, 0));
    digitalWrite(B, bitRead(row, 1));
    digitalWrite(C, bitRead(row, 2));
    digitalWrite(D, bitRead(row, 3));
    digitalWrite(R0, 0);
    digitalWrite(G0, 0);
    digitalWrite(B0, 0);
    digitalWrite(R1, 0);
    digitalWrite(G1, 0);
    digitalWrite(B1, 0);

    column = 0;
    while (column < WIDTH)
    {
      temp = cells[row] >> column; // Top half of matrix

      switch (Game_Num)
      {
        case Life_Display:
          break;
        case Snowflake_Display:
          break;
        case Snake_Game:
          if ((column == Snake[0][0]) && (row == Snake[0][1]))
            color = GRN;
          else
            color = RED;
          break;
        case Pong_Game:
          if ((column == Ball[0]) && (row == Ball[1]))
            color = GRN;
          else
            color = RED;
          break;
        case Tetris_Game:
          // Set shape to green, lines to red
          if ((column == Shape[0][0]) && (row == Shape[0][1]))
            color = GRN;
          else if ((column == Shape[1][0]) && (row == Shape[1][1]))
            color = GRN;
          else if ((column == Shape[2][0]) && (row == Shape[2][1]))
            color = GRN;
          else if ((column == Shape[3][0]) && (row == Shape[3][1]))
            color = GRN;
          else
            color = RED;
          break;
      }

      switch (color) {
        case RED:
          digitalWrite(R0, temp & 1);
          digitalWrite(G0, 0);
          break;
        case GRN:
          digitalWrite(G0, temp & 1);
          digitalWrite(R0, 0);
          break;
        case BLU:
          digitalWrite(B0, temp & 1);
          break;
        case YEL:
          digitalWrite(R0, temp & 1);
          digitalWrite(G0, temp & 1);
          break;
        case MAG:
          digitalWrite(R0, temp & 1);
          digitalWrite(B0, temp & 1);
          break;
        case CYN:
          digitalWrite(B0, temp & 1);
          digitalWrite(G0, temp & 1);
          break;
      }

      temp = cells[row + HEIGHT / 2] >> column; // Bottom half of matrix

      switch (Game_Num)
      {
        case Life_Display:
          break;
        case Snowflake_Display:
          break;
        case Snake_Game:
          if ((column == Snake[0][0]) && ((row + (HEIGHT / 2)) == Snake[0][1]))
            color = GRN;
          else
            color = RED;
          break;
        case Pong_Game:
          if ((column == Ball[0]) && ((row + (HEIGHT / 2)) == Ball[1]))
            color = GRN;
          else
            color = RED;
          break;
        case Tetris_Game:
          // Set shape to green, lines to red
          if ((column == Shape[0][0]) && ((row + (HEIGHT / 2)) == Shape[0][1]))
            color = GRN;
          else if ((column == Shape[1][0]) && ((row + (HEIGHT / 2)) == Shape[1][1]))
            color = GRN;
          else if ((column == Shape[2][0]) && ((row + (HEIGHT / 2)) == Shape[2][1]))
            color = GRN;
          else if ((column == Shape[3][0]) && ((row + (HEIGHT / 2)) == Shape[3][1]))
            color = GRN;
          else
            color = RED;
          break;
      }

      switch (color) {
        case RED:
          digitalWrite(R1, temp & 1);
          digitalWrite(G1, 0);
          break;
        case GRN:
          digitalWrite(G1, temp & 1);
          digitalWrite(R1, 0);
          break;
        case BLU:
          digitalWrite(B1, temp & 1);
          break;
        case YEL:
          digitalWrite(R1, temp & 1);
          digitalWrite(G1, temp & 1);
          break;
        case MAG:
          digitalWrite(R1, temp & 1);
          digitalWrite(B1, temp & 1);
          break;
        case CYN:
          digitalWrite(B1, temp & 1);
          digitalWrite(G1, temp & 1);
          break;
      }

      digitalWrite(CLK, HIGH);
      digitalWrite(CLK, LOW);
      column++;
    }
    digitalWrite(LAT, HIGH);
    digitalWrite(LAT, LOW);

    digitalWrite(OE, LOW);
    row++;
    delayMicroseconds(update_delay);
    digitalWrite(OE, HIGH);
  }
}

void Display_Digit(byte value, byte location) {
  byte x;

  if (location == 1)
    x = 7;
  else if (location == 2)
    x = 13;
  else
    x = 19;

  switch (value) {
    case 0:
      bitSet(cells[18], x + 1);
      bitSet(cells[18], x + 2);
      bitSet(cells[18], x + 3);
      bitSet(cells[19], x);
      bitSet(cells[19], x + 4);
      bitSet(cells[20], x);
      bitSet(cells[20], x + 4);
      bitSet(cells[21], x);
      bitSet(cells[21], x + 4);
      bitSet(cells[22], x);
      bitSet(cells[22], x + 4);
      bitSet(cells[23], x);
      bitSet(cells[23], x + 4);
      bitSet(cells[24], x + 1);
      bitSet(cells[24], x + 2);
      bitSet(cells[24], x + 3);
      break;
    case 1:
      bitSet(cells[18], x + 2);
      bitSet(cells[19], x + 1);
      bitSet(cells[19], x + 2);
      bitSet(cells[20], x + 2);
      bitSet(cells[21], x + 2);
      bitSet(cells[22], x + 2);
      bitSet(cells[23], x + 2);
      bitSet(cells[24], x + 1);
      bitSet(cells[24], x + 2);
      bitSet(cells[24], x + 3);
      break;
    case 2:
      bitSet(cells[18], x + 1);
      bitSet(cells[18], x + 2);
      bitSet(cells[18], x + 3);
      bitSet(cells[19], x);
      bitSet(cells[19], x + 4);
      bitSet(cells[20], x + 4);
      bitSet(cells[21], x + 3);
      bitSet(cells[22], x + 2);
      bitSet(cells[23], x + 1);
      bitSet(cells[24], x);
      bitSet(cells[24], x + 1);
      bitSet(cells[24], x + 2);
      bitSet(cells[24], x + 3);
      bitSet(cells[24], x + 4);
      break;
    case 3:
      bitSet(cells[18], x + 1);
      bitSet(cells[18], x + 2);
      bitSet(cells[18], x + 3);
      bitSet(cells[19], x);
      bitSet(cells[19], x + 4);
      bitSet(cells[20], x + 4);
      bitSet(cells[21], x + 3);
      bitSet(cells[21], x + 2);
      bitSet(cells[22], x + 4);
      bitSet(cells[23], x);
      bitSet(cells[23], x + 4);
      bitSet(cells[24], x + 1);
      bitSet(cells[24], x + 2);
      bitSet(cells[24], x + 3);
      break;
    case 4:
      bitSet(cells[18], x + 3);
      bitSet(cells[19], x + 2);
      bitSet(cells[19], x + 3);
      bitSet(cells[20], x + 1);
      bitSet(cells[20], x + 3);
      bitSet(cells[21], x);
      bitSet(cells[21], x + 3);
      bitSet(cells[22], x);
      bitSet(cells[22], x + 1);
      bitSet(cells[22], x + 2);
      bitSet(cells[22], x + 3);
      bitSet(cells[22], x + 4);
      bitSet(cells[23], x + 3);
      bitSet(cells[24], x + 3);
      break;
    case 5:
      bitSet(cells[18], x);
      bitSet(cells[18], x + 1);
      bitSet(cells[18], x + 2);
      bitSet(cells[18], x + 3);
      bitSet(cells[18], x + 4);
      bitSet(cells[19], x);
      bitSet(cells[20], x);
      bitSet(cells[21], x);
      bitSet(cells[21], x + 1);
      bitSet(cells[21], x + 2);
      bitSet(cells[21], x + 3);
      bitSet(cells[22], x + 4);
      bitSet(cells[23], x + 4);
      bitSet(cells[24], x);
      bitSet(cells[24], x + 1);
      bitSet(cells[24], x + 2);
      bitSet(cells[24], x + 3);
      break;
    case 6:
      bitSet(cells[18], x + 1);
      bitSet(cells[18], x + 2);
      bitSet(cells[18], x + 3);
      bitSet(cells[19], x);
      bitSet(cells[20], x);
      bitSet(cells[21], x);
      bitSet(cells[21], x + 1);
      bitSet(cells[21], x + 2);
      bitSet(cells[21], x + 3);
      bitSet(cells[22], x);
      bitSet(cells[22], x + 4);
      bitSet(cells[23], x);
      bitSet(cells[23], x + 4);
      bitSet(cells[24], x + 1);
      bitSet(cells[24], x + 2);
      bitSet(cells[24], x + 3);
      break;
    case 7:
      bitSet(cells[18], x);
      bitSet(cells[18], x + 1);
      bitSet(cells[18], x + 2);
      bitSet(cells[18], x + 3);
      bitSet(cells[18], x + 4);
      bitSet(cells[19], x + 4);
      bitSet(cells[20], x + 3);
      bitSet(cells[21], x + 2);
      bitSet(cells[22], x + 2);
      bitSet(cells[23], x + 2);
      bitSet(cells[24], x + 2);
      break;
    case 8:
      bitSet(cells[18], x + 1);
      bitSet(cells[18], x + 2);
      bitSet(cells[18], x + 3);
      bitSet(cells[19], x);
      bitSet(cells[19], x + 4);
      bitSet(cells[20], x);
      bitSet(cells[20], x + 4);
      bitSet(cells[21], x + 1);
      bitSet(cells[21], x + 2);
      bitSet(cells[21], x + 3);
      bitSet(cells[22], x);
      bitSet(cells[22], x + 4);
      bitSet(cells[23], x);
      bitSet(cells[23], x + 4);
      bitSet(cells[24], x + 1);
      bitSet(cells[24], x + 2);
      bitSet(cells[24], x + 3);
      break;
    case 9:
      bitSet(cells[18], x + 1);
      bitSet(cells[18], x + 2);
      bitSet(cells[18], x + 3);
      bitSet(cells[19], x);
      bitSet(cells[19], x + 4);
      bitSet(cells[20], x);
      bitSet(cells[20], x + 4);
      bitSet(cells[21], x + 1);
      bitSet(cells[21], x + 2);
      bitSet(cells[21], x + 3);
      bitSet(cells[22], x + 4);
      bitSet(cells[23], x + 4);
      bitSet(cells[24], x + 1);
      bitSet(cells[24], x + 2);
      bitSet(cells[24], x + 3);
      break;
  }
}

// Interrupt-on-change handler
ISR (PCINT0_vect) {
  switch (Game_Num)
  {
    case Snake_Game:
      Snake_ISR0();
      break;
    case Pong_Game:
      break;
    case Tetris_Game:
      Tetris_ISR0();
      break;
  }
}

// Interrupt-on-change handler
ISR (PCINT1_vect) {
  switch (Game_Num)
  {
    case Snake_Game:
      Snake_ISR1();
      break;
    case Tetris_Game:
      Tetris_ISR1();
      break;
  }
}
